#include "dynamixel_master_uart_dma.h"
#include "motion_manager.h"
#include "stm32_time.h"

#define     USART                    USART2
#define     USART_CLK                RCC_APB1Periph_USART2
#define     USART_CLK_INIT           RCC_APB1PeriphClockCmd
#define     USART_IRQn               USART2_IRQn
#define     USART_IRQHandler         USART2_IRQHandler

#define     USART_TX_PIN             GPIO_Pin_2                
#define     USART_TX_GPIO_PORT       GPIOA 
#define     USART_TX_GPIO_CLK        RCC_AHB1Periph_GPIOA
#define     USART_TX_SOURCE          GPIO_PinSource2
#define     USART_TX_AF              GPIO_AF_USART2

#define     USART_RX_PIN             GPIO_Pin_3
#define     USART_RX_GPIO_PORT       GPIOA
#define     USART_RX_GPIO_CLK        RCC_AHB1Periph_GPIOA
#define     USART_RX_SOURCE          GPIO_PinSource3
#define     USART_RX_AF              GPIO_AF_USART2

#define     USART_TX_EN_PIN          GPIO_Pin_1
#define     USART_TX_EN_GPIO_PORT    GPIOA
#define     USART_TX_EN_GPIO_CLK     RCC_AHB1Periph_GPIOA
#define     USART_TX_EN_SOURCE       GPIO_PinSource1

#define     USART_RX_EN_PIN          GPIO_Pin_0
#define     USART_RX_EN_GPIO_PORT    GPIOA
#define     USART_RX_EN_GPIO_CLK     RCC_AHB1Periph_GPIOA
#define     USART_RX_EN_SOURCE       GPIO_PinSource0

/* DMA configuration */
#define     USART_DR_ADDRESS         ((uint32_t)USART2 + 0x04) 
#define     USART_DMA                DMA1
#define     USART_DMA_CLK            RCC_AHB1Periph_DMA1

#define     USART_TX_DMA_CHANNEL     DMA_Channel_4
#define     USART_TX_DMA_STREAM      DMA1_Stream6
#define     USART_TX_DMA_FLAG_FEIF   DMA_FLAG_FEIF6
#define     USART_TX_DMA_FLAG_DMEIF  DMA_FLAG_DMEIF6
#define     USART_TX_DMA_FLAG_TEIF   DMA_FLAG_TEIF6
#define     USART_TX_DMA_FLAG_HTIF   DMA_FLAG_HTIF6
#define     USART_TX_DMA_FLAG_TCIF   DMA_FLAG_TCIF6

#define     USART_RX_DMA_CHANNEL     DMA_Channel_4
#define     USART_RX_DMA_STREAM      DMA1_Stream5
#define     USART_RX_DMA_FLAG_FEIF   DMA_FLAG_FEIF5
#define     USART_RX_DMA_FLAG_DMEIF  DMA_FLAG_DMEIF5
#define     USART_RX_DMA_FLAG_TEIF   DMA_FLAG_TEIF5
#define     USART_RX_DMA_FLAG_HTIF   DMA_FLAG_HTIF5
#define     USART_RX_DMA_FLAG_TCIF   DMA_FLAG_TCIF5

#define     USART_DMA_TX_IRQn        DMA1_Stream6_IRQn
#define     USART_DMA_RX_IRQn        DMA1_Stream5_IRQn
#define     USART_DMA_TX_IRQHandler  DMA1_Stream6_IRQHandler
#define     USART_DMA_RX_IRQHandler  DMA1_Stream5_IRQHandler

#define     POWER_GPIO_CLK           RCC_AHB1Periph_GPIOB
#define     POWER_PIN                GPIO_Pin_2               
#define     POWER_GPIO_PORT          GPIOB                       
#define     POWER_SOURCE             GPIO_PinSource2


#define     MAX_BUFFER_LEN           1024

// private variables
uint8_t dyn_master_version;
uint16_t dyn_master_timeout;// answer reception timeout
// input buffer
uint8_t dyn_master_rx_buffer[MAX_BUFFER_LEN];
// output buffer
uint8_t dyn_master_tx_buffer[MAX_BUFFER_LEN];
// instruction packet ready flag
volatile uint8_t dyn_master_packet_ready;
// sending status packet flag
volatile uint8_t dyn_master_sending_packet;
// no answer for sync write operation
uint8_t dyn_master_no_answer;
// power enabled variable
uint8_t dyn_master_power_enabled;
// number of packets to receive 
uint8_t dyn_master_num_read_packets;
uint8_t dyn_master_packet_length;
// DMA initialization data structures
DMA_InitTypeDef  DMA_TX_InitStructure;
DMA_InitTypeDef  DMA_RX_InitStructure;

// private functions
void dyn_master_reset(void)
{
  /* Reset reception */
  USART_DMACmd(USART, USART_DMAReq_Rx, DISABLE);
  DMA_Cmd(USART_RX_DMA_STREAM,DISABLE);
  /* wait until the transaction ends */
  DMA_ClearFlag(USART_RX_DMA_STREAM,USART_RX_DMA_FLAG_TCIF);
  DMA_ClearITPendingBit(USART_RX_DMA_STREAM,USART_RX_DMA_FLAG_TCIF);
  /* clear any pending data */
  USART_ReceiveData(USART);
  /* Reset transmission */
  USART_DMACmd(USART, USART_DMAReq_Tx, DISABLE);
  DMA_Cmd(USART_TX_DMA_STREAM,DISABLE);
  DMA_ClearFlag(USART_TX_DMA_STREAM,USART_TX_DMA_FLAG_TCIF);
  DMA_ClearITPendingBit(USART_TX_DMA_STREAM,USART_TX_DMA_FLAG_TCIF);
  USART_ITConfig(USART, USART_IT_TC, DISABLE);
  /* Reset variables */
  dyn_master_packet_ready=0x00;
  dyn_master_sending_packet=0x00;
}

uint8_t dyn_master_wait_sending(void)
{
  uint16_t timeout_left=dyn_master_timeout*1000;

  while(dyn_master_sending_packet)
  {
    delay_us(100);
    timeout_left-=100;
    if(timeout_left<=0)
    {
      dyn_master_reset();
      return DYN_TIMEOUT;
    }
  }

  return DYN_SUCCESS;
}

uint8_t dyn_master_send(void)
{
  uint8_t error;

  dyn_master_version=1;
  // wait until any previous transmission ends
  if((error=dyn_master_wait_sending())!=DYN_SUCCESS)
    return error;
  // set the DMA transfer
  DMA_SetCurrDataCounter(USART_TX_DMA_STREAM,dyn_get_length(dyn_master_tx_buffer)+4);
  DMA_Cmd(USART_TX_DMA_STREAM,ENABLE);
  USART_ClearFlag(USART,USART_FLAG_TC);
  USART_ClearITPendingBit(USART,USART_FLAG_TC);
  USART_DMACmd(USART, USART_DMAReq_Tx, ENABLE);
  dyn_master_sending_packet=0x01;
  return error;

}

uint8_t dyn2_master_send(void)
{
  uint8_t error;

  dyn_master_version=2;
  // wait until any previous transmission ends
  if((error=dyn_master_wait_sending())!=DYN_SUCCESS)
    return error;
  // set the DMA transfer
  DMA_SetCurrDataCounter(USART_TX_DMA_STREAM,dyn2_get_length(dyn_master_tx_buffer)+7);
  DMA_Cmd(USART_TX_DMA_STREAM,ENABLE);
  USART_ClearFlag(USART,USART_FLAG_TC);
  USART_ClearITPendingBit(USART,USART_FLAG_TC);
  USART_DMACmd(USART, USART_DMAReq_Tx, ENABLE);
  dyn_master_sending_packet=0x01;
  return error;
}

uint8_t dyn_master_receive(void)
{
  uint16_t timeout_left=dyn_master_timeout*1000;

  // wait for the status packet
  while(!dyn_master_packet_ready)
  {
    delay_us(100);
    timeout_left-=100;
    if(timeout_left<=0)
    {
      dyn_master_reset();
      return DYN_TIMEOUT;
    }
  }
  dyn_master_packet_ready=0x00;
  // check the input packet checksum
  if(dyn_check_checksum(dyn_master_rx_buffer)==0xFF)
    return dyn_get_status_error(dyn_master_rx_buffer);
  else
  {
    dyn_master_reset();
    return DYN_CHECKSUM_ERROR;
  }
}

uint8_t dyn2_master_receive(void)
{
  uint16_t timeout_left=dyn_master_timeout*1000;

  // wait for the status packet
  while(!dyn_master_packet_ready)
  {
    delay_us(100);
    timeout_left-=100;
    if(timeout_left<=0)
    {
      dyn_master_reset();
      return DYN_TIMEOUT;
    }
  }
  dyn_master_packet_ready=0x00;
  // check the input packet checksum
  if(dyn2_check_checksum(dyn_master_rx_buffer)==0x01)
    return dyn2_get_status_error(dyn_master_rx_buffer);
  else
  {
    dyn_master_reset();
    return DYN_CHECKSUM_ERROR;
  }
}

void dyn_master_enable_tx(void)
{
  GPIO_ResetBits(USART_RX_EN_GPIO_PORT, USART_RX_EN_PIN);
  GPIO_SetBits(USART_TX_EN_GPIO_PORT, USART_TX_EN_PIN);
}

void dyn_master_enable_rx(void)
{
  GPIO_ResetBits(USART_TX_EN_GPIO_PORT, USART_TX_EN_PIN);
  GPIO_SetBits(USART_RX_EN_GPIO_PORT, USART_RX_EN_PIN);
}

uint8_t dyn_master_read(uint8_t id,uint8_t address,uint8_t length,uint8_t *data)
{
  uint8_t error;

  /* generate the read packet */
  dyn_init_read_packet(dyn_master_tx_buffer,id,address,length);
  dyn_master_num_read_packets=0x01;
  // enable transmission
  dyn_master_enable_tx();
  // send the data
  if((error=dyn_master_send())!=DYN_SUCCESS)
  {
    dyn_master_enable_rx();
    return error;
  }
  // wait for the transmission to end
  if((error=dyn_master_wait_sending())!=DYN_SUCCESS)
  {
    dyn_master_enable_rx();
    return error;
  }
  dyn_master_enable_rx();
  // wait for the replay within the given timeout
  error=dyn_master_receive();
  if(error==DYN_NO_ERROR)
    dyn_get_read_status_data(dyn_master_rx_buffer,data);// read the data from the status packet
  else
    dyn_master_reset();
  return error;
}

uint8_t dyn2_master_read(uint8_t id,uint16_t address,uint16_t length,uint8_t *data)
{
  uint8_t error;

  /* generate the read packet */
  dyn2_init_read_packet(dyn_master_tx_buffer,id,address,length);
  dyn_master_num_read_packets=0x01;
  // enable transmission
  dyn_master_enable_tx();
  // send the data
  if((error=dyn2_master_send())!=DYN_SUCCESS)
  {
    dyn_master_enable_rx();
    return error;
  }
  // wait for the transmission to end
  if((error=dyn_master_wait_sending())!=DYN_SUCCESS)
  {
    dyn_master_enable_rx();
    return error;
  }
  dyn_master_enable_rx();
  // wait for the replay within the given timeout
  error=dyn2_master_receive();
  if(error==DYN_NO_ERROR)
    dyn2_get_read_status_data(dyn_master_rx_buffer,data);// read the data from the status packet
  else
    dyn_master_reset();
  return error;
}

uint8_t dyn_master_write(uint8_t id, uint8_t address, uint8_t length, uint8_t *data)
{
  uint8_t error;

  // generate the read packet for the desired device
  dyn_init_write_packet(dyn_master_tx_buffer,id,address,length,data);
  dyn_master_num_read_packets=0x01;
  // enable transmission
  dyn_master_enable_tx();
  // send the data
  if((error=dyn_master_send())!=DYN_SUCCESS)
  {
    dyn_master_enable_rx();
    return error;
  }
  // wait for the transmission to end
  if((error=dyn_master_wait_sending())!=DYN_SUCCESS)
  {
    dyn_master_enable_rx();
    return error;
  }
  dyn_master_enable_rx();
  // wait for the replay within the given timeout
  error=dyn_master_receive();
  if(error!=DYN_NO_ERROR)
    dyn_master_reset();
  return error;
}

uint8_t dyn2_master_write(uint8_t id, uint16_t address, uint16_t length, uint8_t *data)
{
  uint8_t error;

  // generate the read packet for the desired device
  dyn2_init_write_packet(dyn_master_tx_buffer,id,address,length,data);
  dyn_master_num_read_packets=0x01;
  // enable transmission
  dyn_master_enable_tx();
  // send the data
  if((error=dyn2_master_send())!=DYN_SUCCESS)
  {
    dyn_master_enable_rx();
    return error;
  }
  // wait for the transmission to end
  if((error=dyn_master_wait_sending())!=DYN_SUCCESS)
  {
    dyn_master_enable_rx();
    return error;
  }
  dyn_master_enable_rx();
  // wait for the replay within the given timeout
  error=dyn2_master_receive();
  if(error!=DYN_NO_ERROR)
    dyn_master_reset();
  return error;
}

// interrupt handlers
void USART_IRQHandler(void)
{
  static uint8_t num_bytes=0;
  uint8_t data;

  if(USART_GetITStatus(USART, USART_IT_RXNE) != RESET)
  {
    USART_ClearFlag(USART,USART_FLAG_RXNE);
    data=USART_ReceiveData(USART);
    switch(num_bytes)
    {
      case 0: if(data==0xFF)
              {
                dyn_master_rx_buffer[num_bytes]=data;
                num_bytes++;
              }
              break;
      case 1: if(data==0xFF)
              {
                dyn_master_rx_buffer[num_bytes]=data;
                num_bytes++;
              }
              else num_bytes--;
              break;
      case 2: if(data!=0xFF)
              {
                dyn_master_rx_buffer[num_bytes]=data;
                num_bytes++;
              }
              break;
      case 3: dyn_master_rx_buffer[num_bytes]=data;
              num_bytes=0;
              /* disable USART RX interrupts */
              USART_ITConfig(USART,USART_IT_RXNE,DISABLE);
              /* enable DMA RX */
              DMA_RX_InitStructure.DMA_BufferSize = data;
              DMA_RX_InitStructure.DMA_Memory0BaseAddr = (uint32_t)&dyn_master_rx_buffer[dyn_master_packet_length+4];
              dyn_master_packet_length+=data+4;
              DMA_Init(USART_RX_DMA_STREAM,&DMA_RX_InitStructure);
              DMA_Cmd(USART_RX_DMA_STREAM,ENABLE);
              USART_DMACmd(USART, USART_DMAReq_Rx, ENABLE);
              break;
      default: break;
    }
  }
  else if(USART_GetITStatus(USART, USART_IT_TC) != RESET)
  {
    USART_ClearFlag(USART,USART_FLAG_TC);
    USART_ITConfig(USART, USART_IT_TC, DISABLE);
    dyn_master_sending_packet=0x00;
    if(dyn_master_no_answer)
      dyn_master_no_answer=0x00;
    else
    {
      /* enable RX interrupts */
      USART_ITConfig(USART, USART_IT_RXNE, ENABLE);
    }
  }
}

void USART_DMA_TX_IRQHandler(void)
{
  DMA_Cmd(USART_TX_DMA_STREAM,DISABLE);
  DMA_ClearFlag(USART_TX_DMA_STREAM,USART_TX_DMA_FLAG_TCIF);
  DMA_ClearITPendingBit(USART_TX_DMA_STREAM,USART_TX_DMA_FLAG_TCIF);
  USART_DMACmd(USART, USART_DMAReq_Tx, DISABLE);
  USART_ITConfig(USART, USART_IT_TC, ENABLE);
}

void USART_DMA_RX_IRQHandler(void)
{
  DMA_Cmd(USART_RX_DMA_STREAM,DISABLE);
  DMA_ClearFlag(USART_RX_DMA_STREAM,USART_RX_DMA_FLAG_TCIF);
  DMA_ClearITPendingBit(USART_RX_DMA_STREAM,USART_RX_DMA_FLAG_TCIF);
  USART_DMACmd(USART, USART_DMAReq_Rx, DISABLE);
  dyn_master_num_read_packets--;
  if(dyn_master_num_read_packets==0x00)
  {
    dyn_master_packet_length=0x00;
    dyn_master_packet_ready=0x01;
  }
  else
  {
    /* enable RX interrupts */
    USART_ITConfig(USART, USART_IT_RXNE, ENABLE);
  }
}

// dynamixel master functions
void dyn_master_init(void)
{
  uint16_t i;
  GPIO_InitTypeDef GPIO_InitStructure;
  NVIC_InitTypeDef NVIC_InitStructure;
  USART_InitTypeDef USART_InitStructure;

  /* Enable GPIO clock */
  USART_CLK_INIT(USART_CLK, ENABLE);
  RCC_AHB1PeriphClockCmd(USART_TX_GPIO_CLK | USART_RX_GPIO_CLK | USART_TX_EN_GPIO_CLK | USART_RX_EN_GPIO_CLK, ENABLE);
  RCC_AHB1PeriphClockCmd(POWER_GPIO_CLK, ENABLE);
  // configure the GPIO pins

  /* Connect USART pins to AF7 */
  GPIO_PinAFConfig(USART_TX_GPIO_PORT, USART_TX_SOURCE, USART_TX_AF);
  GPIO_PinAFConfig(USART_RX_GPIO_PORT, USART_RX_SOURCE, USART_RX_AF);

  /* Configure USART Tx and Rx as alternate function push-pull */
  GPIO_InitStructure.GPIO_Mode = GPIO_Mode_AF;
  GPIO_InitStructure.GPIO_Speed = GPIO_Speed_100MHz;
  GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
  GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
  GPIO_InitStructure.GPIO_Pin = USART_TX_PIN;
  GPIO_Init(USART_TX_GPIO_PORT, &GPIO_InitStructure);

  GPIO_InitStructure.GPIO_Pin = USART_RX_PIN;
  GPIO_Init(USART_RX_GPIO_PORT, &GPIO_InitStructure);

  /* configure the control pins */
  GPIO_InitStructure.GPIO_Pin = USART_TX_EN_PIN;
  GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
  GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
  GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
  GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
  GPIO_Init(USART_TX_EN_GPIO_PORT, &GPIO_InitStructure);
  GPIO_InitStructure.GPIO_Pin = USART_RX_EN_PIN;
  GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
  GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
  GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
  GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
  GPIO_Init(USART_RX_EN_GPIO_PORT, &GPIO_InitStructure);

  dyn_master_enable_rx();

  dyn_master_timeout=0x00;
  // initialize the buffers
  for(i=0;i<MAX_BUFFER_LEN;i++)
  {
    dyn_master_rx_buffer[i]=0x00;
    dyn_master_tx_buffer[i]=0x00;
  }
  // initialize the flags
  dyn_master_packet_ready=0x00;
  dyn_master_sending_packet=0x00;
  dyn_master_no_answer=0x00;
  dyn_master_num_read_packets=0x00;
  dyn_master_packet_length=0x00;
  dyn_master_version=1;

  USART_DeInit(USART);
  USART_StructInit(&USART_InitStructure);
  // configure the serial port
  USART_InitStructure.USART_BaudRate = 1000000;
  USART_InitStructure.USART_WordLength = USART_WordLength_8b;
  USART_InitStructure.USART_StopBits = USART_StopBits_1;
  USART_InitStructure.USART_Parity = USART_Parity_No;
  USART_InitStructure.USART_HardwareFlowControl = USART_HardwareFlowControl_None;
  USART_InitStructure.USART_Mode = USART_Mode_Rx | USART_Mode_Tx;
  USART_Init(USART, &USART_InitStructure);
  NVIC_InitStructure.NVIC_IRQChannel = USART_IRQn;
  NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 0;
  NVIC_InitStructure.NVIC_IRQChannelSubPriority = 1;
  NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
  NVIC_Init(&NVIC_InitStructure);

  USART_ITConfig(USART, USART_IT_RXNE | 
                        USART_IT_ORE | 
                        USART_IT_TXE | 
                        USART_IT_CTS | 
                        USART_IT_LBD | 
                        USART_IT_IDLE | 
                        USART_IT_PE | 
                        USART_IT_ERR | 
                        USART_IT_TC, DISABLE);

  /* Clear all USART interrupt flags */
  USART_ClearFlag(USART,USART_FLAG_RXNE |
                        USART_FLAG_ORE |
                        USART_FLAG_TXE |
                        USART_FLAG_CTS |
                        USART_FLAG_LBD |
                        USART_FLAG_IDLE |
                        USART_FLAG_PE |
                        USART_FLAG_TC);

  /* Clear all USART interrupt pending bits */
  USART_ClearITPendingBit(USART, USART_FLAG_RXNE |
                                 USART_FLAG_ORE |
                                 USART_FLAG_TXE |
                                 USART_FLAG_CTS |
                                 USART_FLAG_LBD |
                                 USART_FLAG_IDLE |
                                 USART_FLAG_PE |
                                 USART_FLAG_TC);

  // configure the DMA channels
  /* Configure TX DMA */
  RCC_AHB1PeriphClockCmd(USART_DMA_CLK, ENABLE);
  DMA_DeInit(USART_TX_DMA_STREAM);
  DMA_TX_InitStructure.DMA_FIFOMode = DMA_FIFOMode_Disable;
  DMA_TX_InitStructure.DMA_FIFOThreshold = DMA_FIFOThreshold_1QuarterFull;
  DMA_TX_InitStructure.DMA_MemoryBurst = DMA_MemoryBurst_Single;
  DMA_TX_InitStructure.DMA_MemoryDataSize = DMA_MemoryDataSize_Byte;
  DMA_TX_InitStructure.DMA_MemoryInc = DMA_MemoryInc_Enable;
  DMA_TX_InitStructure.DMA_Mode = DMA_Mode_Normal;
  DMA_TX_InitStructure.DMA_PeripheralBaseAddr = (uint32_t) (&(USART->DR));
  DMA_TX_InitStructure.DMA_PeripheralBurst = DMA_PeripheralBurst_Single;
  DMA_TX_InitStructure.DMA_PeripheralDataSize = DMA_PeripheralDataSize_Byte;
  DMA_TX_InitStructure.DMA_PeripheralInc = DMA_PeripheralInc_Disable;
  DMA_TX_InitStructure.DMA_Priority = DMA_Priority_High;
  DMA_TX_InitStructure.DMA_Channel = USART_TX_DMA_CHANNEL;
  DMA_TX_InitStructure.DMA_DIR = DMA_DIR_MemoryToPeripheral;
  DMA_TX_InitStructure.DMA_Memory0BaseAddr = (uint32_t)dyn_master_tx_buffer;
  DMA_Init(USART_TX_DMA_STREAM,&DMA_TX_InitStructure);
  /* initialize DMA interrupts */
  NVIC_InitStructure.NVIC_IRQChannel = USART_DMA_TX_IRQn;
  NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 1;
  NVIC_InitStructure.NVIC_IRQChannelSubPriority = 2;
  NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
  NVIC_Init(&NVIC_InitStructure);
  DMA_ITConfig(USART_TX_DMA_STREAM,DMA_IT_TC,ENABLE);
  DMA_ITConfig(USART_TX_DMA_STREAM,DMA_IT_HT | DMA_IT_TE | DMA_IT_FE | DMA_IT_DME,DISABLE);

  /* Configure RX DMA */
  DMA_DeInit(USART_RX_DMA_STREAM);
  DMA_RX_InitStructure.DMA_FIFOMode = DMA_FIFOMode_Disable;
  DMA_RX_InitStructure.DMA_FIFOThreshold = DMA_FIFOThreshold_Full;
  DMA_RX_InitStructure.DMA_MemoryBurst = DMA_MemoryBurst_Single;
  DMA_RX_InitStructure.DMA_MemoryDataSize = DMA_MemoryDataSize_Byte;
  DMA_RX_InitStructure.DMA_MemoryInc = DMA_MemoryInc_Enable;
  DMA_RX_InitStructure.DMA_Mode = DMA_Mode_Normal;
  DMA_RX_InitStructure.DMA_PeripheralBaseAddr = (uint32_t) (&(USART->DR));
  DMA_RX_InitStructure.DMA_PeripheralBurst = DMA_PeripheralBurst_Single;
  DMA_RX_InitStructure.DMA_PeripheralDataSize = DMA_PeripheralDataSize_Byte;
  DMA_RX_InitStructure.DMA_PeripheralInc = DMA_PeripheralInc_Disable;
  DMA_RX_InitStructure.DMA_Priority = DMA_Priority_High;
  DMA_RX_InitStructure.DMA_Channel = USART_RX_DMA_CHANNEL;
  DMA_RX_InitStructure.DMA_DIR = DMA_DIR_PeripheralToMemory;
  DMA_RX_InitStructure.DMA_Memory0BaseAddr = (uint32_t)&dyn_master_rx_buffer[4];
  DMA_Init(USART_RX_DMA_STREAM,&DMA_RX_InitStructure);
  /* initialize DMA interrupts */
  NVIC_InitStructure.NVIC_IRQChannel = USART_DMA_RX_IRQn;
  NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 1;
  NVIC_InitStructure.NVIC_IRQChannelSubPriority = 3;
  NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
  NVIC_Init(&NVIC_InitStructure);
  DMA_ITConfig(USART_RX_DMA_STREAM,DMA_IT_TC,ENABLE);
  DMA_ITConfig(USART_RX_DMA_STREAM,DMA_IT_HT | DMA_IT_TE | DMA_IT_FE | DMA_IT_DME,DISABLE);
  
  /* Enable the USART2 */
  USART_Cmd(USART, ENABLE);

  // initialize the power enable gpio
  GPIO_InitStructure.GPIO_Pin = POWER_PIN;
  GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
  GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
  GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
  GPIO_InitStructure.GPIO_PuPd  = GPIO_PuPd_NOPULL;
  GPIO_Init(POWER_GPIO_PORT, &GPIO_InitStructure);

  dyn_master_disable_power();
}

void dyn_master_set_timeout(uint16_t timeout_ms)
{
  // save the desired timeout value
  dyn_master_timeout=timeout_ms;
}

void dyn_master_scan(uint8_t *num,uint8_t *ids)
{
  uint8_t i;

  for(i=0;i<MANAGER_MAX_NUM_SERVOS;i++)
    ids[0]=0;
  *num=0;
  for(i=0;i<254;i++)
  {
    if(dyn_master_ping(i)==DYN_NO_ERROR)// the device exists
    {
      ids[*num]=i;
      (*num)++;
    }
  }
}

void dyn2_master_scan(uint8_t *num,uint8_t *ids)
{
  uint8_t i;

  for(i=0;i<MANAGER_MAX_NUM_SERVOS;i++)
    ids[0]=0;
  *num=0;
  for(i=0;i<254;i++)
  {
    if(dyn2_master_ping(i)==DYN_NO_ERROR)// the device exists
    {
      ids[*num]=i;
      (*num)++;
    }
  }
}

inline void dyn_master_enable_power(void)
{
  GPIO_SetBits(POWER_GPIO_PORT,POWER_PIN);
  dyn_master_power_enabled=0x01;
}

inline void dyn_master_disable_power(void)
{
  GPIO_ResetBits(POWER_GPIO_PORT,POWER_PIN);
  dyn_master_power_enabled=0x00;
}

uint8_t dyn_master_is_power_enabled(void)
{
  return dyn_master_power_enabled;
}

uint8_t dyn_master_ping(uint8_t id)
{
  uint8_t error;
    
  // generate the ping packet for the desired device
  dyn_init_ping_packet(dyn_master_tx_buffer,id);
  dyn_master_num_read_packets=0x01;
  // enable transmission
  dyn_master_enable_tx();
  // send the data
  if((error=dyn_master_send())!=DYN_SUCCESS)
  { 
    dyn_master_enable_rx();
    return error;
  }
  // wait for the transmission to end
  if((error=dyn_master_wait_sending())!=DYN_SUCCESS)
  {
    dyn_master_enable_rx();
    return error;
  }
  dyn_master_enable_rx();
  // wait for the replay within the given timeout
  error=dyn_master_receive();
  return error;
}

uint8_t dyn2_master_ping(uint8_t id)
{
  uint8_t error;

  // generate the ping packet for the desired device
  dyn2_init_ping_packet(dyn_master_tx_buffer,id);
  dyn_master_num_read_packets=0x01;
  // enable transmission
  dyn_master_enable_tx();
  // send the data
  if((error=dyn2_master_send())!=DYN_SUCCESS)
  {
    dyn_master_enable_rx();
    return error;
  }
  // wait for the transmission to end
  if((error=dyn_master_wait_sending())!=DYN_SUCCESS)
  {
    dyn_master_enable_rx();
    return error;
  }
  dyn_master_enable_rx();
  // wait for the replay within the given timeout
  error=dyn2_master_receive();
  return error;
}

uint8_t dyn_master_read_byte(uint8_t id,uint8_t address,uint8_t *data)
{
  return dyn_master_read(id,address,1,data);
}

uint8_t dyn2_master_read_byte(uint8_t id,uint16_t address,uint8_t *data)
{
  return dyn2_master_read(id,address,1,data);
}

uint8_t dyn_master_read_word(uint8_t id,uint8_t address,uint16_t *data)
{
  uint8_t value[2];
  uint8_t error;

  error=dyn_master_read(id,address,2,value);
  if(error==DYN_NO_ERROR)
    (*data)=value[0]+value[1]*256;

  return error;
}

uint8_t dyn2_master_read_word(uint8_t id,uint16_t address,uint16_t *data)
{
  uint8_t value[2];
  uint8_t error;

  error=dyn2_master_read(id,address,2,value);
  if(error==DYN_NO_ERROR)
    (*data)=value[0]+value[1]*256;

  return error;
}

uint8_t dyn_master_read_table(uint8_t id,uint8_t address,uint8_t length,uint8_t *data)
{
  return dyn_master_read(id,address,length,data);
}

uint8_t dyn2_master_read_table(uint8_t id,uint16_t address,uint16_t length,uint8_t *data)
{
  return dyn2_master_read(id,address,length,data);
}

uint8_t dyn_master_write_byte(uint8_t id, uint8_t address, uint8_t data)
{
  return dyn_master_write(id,address,1,&data);
}

uint8_t dyn2_master_write_byte(uint8_t id, uint16_t address, uint8_t data)
{
  return dyn2_master_write(id,address,1,&data);
}

uint8_t dyn_master_write_word(uint8_t id, uint8_t address, uint16_t data)
{
  uint8_t value[2];
  
  value[0]=data%256;
  value[1]=data/256;
  return dyn_master_write(id,address,2,value);
}

uint8_t dyn2_master_write_word(uint8_t id, uint16_t address, uint16_t data)
{
  uint8_t value[2];

  value[0]=data%256;
  value[1]=data/256;
  return dyn2_master_write(id,address,2,value);
}

uint8_t dyn_master_write_table(uint8_t id, uint8_t address, uint8_t length, uint8_t *data)
{
  return dyn_master_write(id,address,length,data);
}

uint8_t dyn2_master_write_table(uint8_t id, uint16_t address, uint16_t length, uint8_t *data)
{
  return dyn2_master_write(id,address,length,data);
}

uint8_t dyn_master_reg_write(uint8_t id, uint8_t address, uint8_t length, uint8_t *data)
{
  uint8_t error;

  // generate the ping packet for the desired device
  dyn_init_reg_write_packet(dyn_master_tx_buffer,id,address,length,data);
  dyn_master_num_read_packets=0x01;
  // enable transmission
  dyn_master_enable_tx();
  // send the data
  if((error=dyn_master_send())!=DYN_SUCCESS)
  {
    dyn_master_enable_rx();
    return error;
  }
  // wait for the transmission to end
  if((error=dyn_master_wait_sending())!=DYN_SUCCESS)
  {
    dyn_master_enable_rx();
    return error;
  }
  dyn_master_enable_rx();
  // wait for the replay within the given timeout
  error=dyn_master_receive();
  return error;
}

uint8_t dyn2_master_reg_write(uint8_t id, uint16_t address, uint16_t length, uint8_t *data)
{
  uint8_t error;

  // generate the ping packet for the desired device
  dyn2_init_reg_write_packet(dyn_master_tx_buffer,id,address,length,data);
  dyn_master_num_read_packets=0x01;
  // enable transmission
  dyn_master_enable_tx();
  // send the data
  if((error=dyn2_master_send())!=DYN_SUCCESS)
  {
    dyn_master_enable_rx();
    return error;
  }
  // wait for the transmission to end
  if((error=dyn_master_wait_sending())!=DYN_SUCCESS)
  {
    dyn_master_enable_rx();
    return error;
  }
  dyn_master_enable_rx();
  // wait for the replay within the given timeout
  error=dyn2_master_receive();
  return error;
}

uint8_t dyn_master_action(void)
{
  dyn_master_no_answer=0x01;
  // generate the ping packet for the desired device
  dyn_init_action_packet(dyn_master_tx_buffer);
  // enable transmission
  dyn_master_enable_tx();
  // send the data
  return dyn_master_send();
}

uint8_t dyn2_master_action(void)
{
  dyn_master_no_answer=0x01;
  // generate the ping packet for the desired device
  dyn2_init_action_packet(dyn_master_tx_buffer);
  // enable transmission
  dyn_master_enable_tx();
  // send the data
  return dyn2_master_send();
}

uint8_t dyn_master_sync_write(uint8_t num,uint8_t *ids,uint8_t address, uint8_t length, TWriteData *data)
{
  dyn_master_no_answer=0x01;
  // generate the sync write packet 
  dyn_init_sync_write_packet(dyn_master_tx_buffer,num,ids,address,length,data);
  // enable transmission
  dyn_master_enable_tx();
  // send the data
  return dyn_master_send();
}

uint8_t dyn2_master_sync_write(uint8_t num,uint8_t *ids,uint16_t address, uint16_t length, TWriteData *data)
{
  dyn_master_no_answer=0x01;
  // generate the sync write packet 
  dyn2_init_sync_write_packet(dyn_master_tx_buffer,num,ids,address,length,data);
  // enable transmission
  dyn_master_enable_tx();
  // send the data
  return dyn2_master_send();
}

// repeater functions
uint8_t dyn_master_resend_inst_packet(uint8_t *inst_packet, uint8_t *status_packet)
{
  uint8_t error;

  // copy the contents of the instruction packet to the output buffer
  dyn_copy_packet(inst_packet,dyn_master_tx_buffer);
  dyn_master_num_read_packets=0x01;
  // enable transmission
  dyn_master_enable_tx();
  // send the data
  if((error=dyn_master_send())!=DYN_SUCCESS)
  {
    dyn_master_enable_rx();
    return error;
  }
  // wait for the transmission to end
  if((error=dyn_master_wait_sending())!=DYN_SUCCESS)
  {
    dyn_master_enable_rx();
    return error;
  }
  dyn_master_enable_rx();
  // wait for the replay within the given timeout
  error=dyn_master_receive();
  if(error!=DYN_TIMEOUT)
  {
    // copy the contents of the status of the packet to the output buffer
    dyn_copy_packet(dyn_master_rx_buffer,status_packet);
  }

  return error;
}

uint8_t dyn_master_bulk_read(uint8_t num,uint8_t *ids,uint8_t *address,uint8_t *length,TWriteData *data)
{
  uint16_t offset=0;
  uint8_t i,error;

  dyn_master_num_read_packets=num;
  // generate the sync write packet 
  dyn_init_bulk_read_packet(dyn_master_tx_buffer,num,ids,address,length);
  // enable transmission
  dyn_master_enable_tx();
  // send the data
  if((error=dyn_master_send())!=DYN_SUCCESS)
  {
    dyn_master_enable_rx();
    return error;
  }
  // wait for the transmission to end
  if((error=dyn_master_wait_sending())!=DYN_SUCCESS)
  {
    dyn_master_enable_rx();
    return error;
  }
  dyn_master_enable_rx();
  // wait for the replay within the given timeout
  error=dyn_master_receive();
  if(error!=DYN_NO_ERROR)
  {
    dyn_master_reset();
    return error;
  }
  for(i=0;i<num;i++)
  {
    error=dyn_get_status_error(&dyn_master_rx_buffer[offset]);
    if(error!=DYN_NO_ERROR)
      return error;
    offset+=dyn_get_read_status_data(&dyn_master_rx_buffer[offset],data[i].data_addr)+4;
  }

  return DYN_SUCCESS;
}


